import logging

# Silence requests logger
requests_log = logging.getLogger("requests")
requests_log.setLevel(logging.WARNING)

log = logging.getLogger('s3tup')
log.addHandler(logging.NullHandler())

from s3tup.connection import Connection, stats, stats_lock
from s3tup.parse import load_config, parse_config

def s3tup(config, access_key_id=None, secret_access_key=None,
          rsync_only=False):
    log.info('**** s3tup ****')

    config = load_config(config)
    buckets = parse_config(config)

    for b in buckets:
        if b.conn is None:
            b.conn = Connection(access_key_id, secret_access_key)
        b.sync(rsync_only)
    
    log.debug('request totals:')
    with stats_lock:
        for k,v in stats.iteritems():
            if v > 0:
                log.debug('- {}: {}'.format(k,v))


"""
s3tup.parse
~~~~~~~~~~~

This module contains all of the parsing methods used to convert YAML
configuration files into usable s3tup objects.
"""

import logging

import yaml

from s3tup.bucket import Bucket
from s3tup.key import KeyFactory, KeyConfigurator, Key
from s3tup.rsync import Rsync
from s3tup.exception import ConfigParseError, ConfigLoadError

log = logging.getLogger('s3tup.parse')

def load_config(config):
    """Return parseable s3tup configuration from whatever 'config' is.

    Probably the most magic method here, load_config guesses how to convert
    'config' into a regular s3tup config list based on its type.

    Currently supports:

        * config is basestring: load yaml from file at 'config'
        * config is file: load yaml from config.read()
        * config is dict: put into list
        * config is list: pass it on through

    """
    if isinstance(config, basestring):
        try:
            config = file(config)
        except IOError:
            raise ConfigLoadError("Config file '{}' does not exist"
                                  .format(config))

    if isinstance(config, file):
        try:
            config = yaml.load(config)
        except yaml.YAMLError as e:
            raise ConfigLoadError(e)

    if isinstance(config, dict):
        config = [config,]

    return config

# YAML -> python data -> s3tup objects
# Each parse method delegates part of its processing to more specific
# parse methods down the line.

def parse_config(config):
    """Return a list of fully configured Buckets from 'config'."""
    if not isinstance(config, list):
        raise ConfigParseError('Config must be a list')

    buckets = []
    for b in config:
        buckets.append(parse_bucket(b))
    return buckets

def parse_bucket(config):
    """Return a properly configured Bucket from 'config'.

    Note: If access_key_id and secret_access_key are not in the config
    dict, the returned bucket will have a conn of None, *not* a default
    Connection that tries to load from env vars.

    """
    try:
        bucket_name = config.pop('bucket')
    except KeyError:
        raise ConfigParseError("Bucket config must contain field 'bucket'")

    try:
        access_key_id = config.pop('access_key_id')
        secret_access_key = config.pop('secret_access_key')
    except KeyError:
        conn = None
    else:
        conn = Connection(access_key_id, secret_access_key)

    if 'key_config' in config:
        key_factory = parse_key_config(config.pop('key_config'))
    else:
        key_factory = None

    if 'rsync' in config:
        rsync = parse_rsync(config.pop('rsync'))
    else:
        rsync = []
    try:
        return Bucket(conn, bucket_name, key_factory, rsync, **config)
    except TypeError as e:
        raise ConfigParseError(e)

def parse_rsync(config):
    """Return a list of Rsync objects from 'config'

    This has some magic in that it'll accept a string, dict, or list.
    Done to make rsync definitions less verbose when they don't need to be.

    Examples of valid yaml configs for the rsync field:

        rsync: path/to/some/folder            # string

        rsync:
          src: path/to/some/folder            # dict
          delete: true

        rsync:
          - src: path/to/some/folder          # list
            delete: true
          - src: path/to/some/other/folder
            dest: some/prefix/

    Regardless of the input type it returns a list.
    """
    if isinstance(config, basestring):
        config = [{'src': config},]
    elif isinstance(config, dict):
        config = [config,]

    rsync = []
    for r in config:
        rsync.append(parse_rsync_object(r))
    return rsync

def parse_rsync_object(config):
    """Return a properly configured Rsync object from 'config'"""
    matcher, config = extract_matcher(config)
    try:
        return Rsync(matcher=matcher, **config)
    except TypeError as e:
        raise ConfigParseError(e)

def parse_key_config(config):
    """Return a properly configured KeyFactory from 'config'"""
    # Unlike rsync, key_config must be a list
    if not isinstance(config, list):
        raise ConfigParseError('key_config must be a list')

    configurators = []
    for c in config:
        matcher, config = extract_matcher(c)
        configurators.append(parse_key_configurator(c))
    return KeyFactory(configurators)

def parse_key_configurator(config):
    """Return a properly configured KeyConfigurator from 'config'"""
    matcher, config = extract_matcher(config)
    try:
        return KeyConfigurator(matcher=matcher, **config)
    except TypeError as e:
        raise ConfigParseError(e)

def extract_matcher(config):
    """Return (matcher, config) tuple extracted from 'config'

    Extracts the matcher fields from the passed in config dict, creates a
    Matcher object from them, and then returns the Matcher and the config
    dict minus the matcher fields back in a tuple. If none of the matcher
    fields are present, the matcher returned will be None.
    """
    patterns = config.pop('patterns', None)
    ignore_patterns = config.pop('ignore_patterns', None)
    regexes = config.pop('regexes', None)
    ignore_regexes = config.pop('ignore_regexes', None)

    if patterns and ignore_patterns and regexes and ignore_regexes:
        matcher = Matcher(patterns, ignore_patterns, regexes, ignore_regexes)
    else:
        matcher = None
    return matcher, config


import logging
import mimetypes
import os

import s3tup.utils as utils
import s3tup.constants as constants

log = logging.getLogger('s3tup.key')

class KeyFactory(object):
    """
    Basically just a container for KeyConfigurators. make_key will create a
    key based on the input parameters and then run each configurator on it
    sequentially, returning a fully configured key ready for sync.

    """
    def __init__(self, configurators=None):
        self.configurators = configurators or []

    def make_key(self, conn, key_name, bucket_name):
        """Return a properly configured key"""
        key = Key(conn, key_name, bucket_name)
        return self.configure_key(key)

    def configure_key(self, key):
        for c in self.configurators:
            if c.effects_key(key.name):
                key = c.configure_key(key)
        return key


class KeyConfigurator(object):
    def __init__(self, matcher=None, **kwargs):
        self.matcher = matcher
        for k, v in kwargs.iteritems():
            if k in constants.KEY_ATTRS:
                self.__dict__[k] = v
            else:
                raise TypeError("__init__() got an unexpected keyword"
                                " argument '{}'".format(k))

    def effects_key(self, key_name):
        """Return whether this configurator effects key_name"""
        if self.matcher is None:
            return True
        else:
            return self.matcher.matches(key_name)

    def configure_key(self, key):
        """Return the input key with all configurations applied"""
        for attr in constants.KEY_ATTRS:
            if attr in self.__dict__ and attr != 'metadata':
                key.__dict__[attr] = self.__dict__[attr]

        try: key.metadata.update(self.metadata)
        except AttributeError: pass

        return key


class Key(object):
    """
    Encapsulates configuration for a particular s3 key. It has attributes
    (all defined in constants.KEY_ATTRS) that you can set, delete, modify, 
    and then sync to s3 using the sync or rsync methods.

    """
    def __init__(self, conn, name, bucket_name, **kwargs):
        self.conn = conn
        self.name = name
        self.bucket_name = bucket_name

        # Set defaults for required attributes
        self.reduced_redundancy = kwargs.pop('reduced_redundancy', False)
        self.encrypted = kwargs.pop('encrypted', False)
        self.metadata = kwargs.pop('metadata', {})

        for k,v in kwargs.iteritems():
            if k in constants.KEY_ATTRS:
                self.__dict__[k] = v
            else:
                raise TypeError("__init__() got an unexpected keyword"
                                " argument '{}'".format(k))

    def make_request(self, method, params=None, data=None, headers=None):
        """
        Convenience method for self.conn.make_request; has bucket and key
        already filled in.
        """
        return self.conn.make_request(method, self.bucket_name, self.name,
                                      params, data=data, headers=headers)

    @property
    def headers(self):
        """Return the headers associated with this key"""
        headers = {}

        try: headers['x-amz-acl'] = self.canned_acl
        except AttributeError: pass

        if self.reduced_redundancy:
            headers['x-amz-storage-class'] = 'REDUCED_REDUNDANCY'

        if self.encrypted:
            headers['x-amz-server-side-encryption'] = 'AES256'

        for k, v in self.metadata.iteritems():
            headers['x-amz-meta-' + k] = v

        for k in constants.KEY_HEADERS:
            try:
                if self.__dict__[k] is not None:
                    headers[k.replace('_', '-')] = self.__dict__[k]
            except KeyError: pass

        # Guess content-type
        if 'content-type' not in headers:
            content_type_guess = mimetypes.guess_type(self.name)[0]
            if content_type_guess is not None:
                headers['Content-Type'] = content_type_guess

        return headers

    def sync(self):
        """Sync this object's configuration with its respective key on s3"""
        log.info("syncing key '{}'...".format(self.name))

        headers = self.headers
        headers['x-amz-copy-source'] = '/'+self.bucket_name+'/'+self.name
        headers['x-amz-metadata-directive'] = 'REPLACE'

        self.make_request('PUT', headers=headers)
        self.sync_acl()

    def rsync(self, file_like_object):
        """Upload file_like_object to s3 with this object's configuration"""
        log.info("uploading key '{}'...".format(self.name))

        headers = self.headers
        data = file_like_object.read()

        self.make_request('PUT', headers=headers, data=data)
        self.sync_acl()

    def sync_acl(self):
        try: acl = self.acl
        except AttributeError: return False

        if acl is not None:
            self.make_request('PUT', 'acl', data=acl)
        else:
            self.make_request('PUT', 'acl', headers={'x-amz-acl': 'private'})
        
        


"""
s3tup.utils
~~~~~~~~~~~

This module contains various utility functions and classes.
"""

from fnmatch import fnmatch
from base64 import b64encode
import hashlib
import os
import re

class Matcher(object):
    """Matches strings based on patterns

    Internally the matcher object has a set of unix style patterns to match
    and to ignore, and a set of regexes to match and to ignore. Each of
    these are tested against the string 's' in the matches method. Ignore
    has the highest precedence. If both patterns and regexes are empty, the
    object assumes that the match is True (though the ignore sets can still
    overpower this assumption).

    """
    def __init__(self, patterns=None, ignore_patterns=None, regexes=None,
                 ignore_regexes=None):
        self.patterns = patterns or set()
        self.ignore_patterns = ignore_patterns or set()
        self.regexes = regexes or set()
        self.ignore_regexes = ignore_regexes or set()

    def matches(self, s):
        """Return whether this matcher matches string s"""
        # If neither patterns nor regexes is set, match everything
        matched = not self.patterns and not self.regexes
        for pattern in self.patterns:
            if fnmatch(s, pattern):
                matched = True
                break
        if not matched:
            for regex in self.regexes:
                if re.search(regex, s):
                    matched = True
                    break
        if matched:
            for pattern in self.ignore_patterns:
                if fnmatch(s, pattern):
                    return False
        if matched:
            for regex in self.ignore_regexes:
                if re.search(regex, s):
                    return False
        return matched

    def __add__(self, other):
        """Allow combining matchers"""
        patterns = self.patterns | other.patterns
        ignore_patterns = self.ignore_patterns | other.ignore_patterns
        regexes = self.regexes | other.regexes
        ignore_regexes = self.ignore_regexes | other.ignore_regexes

        return Matcher(patterns, ignore_patterns, regexes, ignore_regexes)

    def __iadd__(self, other):
        return self.__add__(other)


def file_md5(f):
    """Return base64 encoded md5 hash of filelike object f"""
    m = hashlib.md5()
    while True:
        buf=f.read(8192)
        if not buf: break
        m.update(buf)
    return b64encode(m.digest()).strip()

def os_walk_relative(src):
    """Return list of all file paths in src relative to src"""
    for path, dirs, files in os.walk(src):
        for f in files:
            full_path = os.path.join(path, f)
            yield os.path.relpath(full_path, src)

def key_diff(before, after):
    """
    Return a dict representing the difference between two runs of
    Bucket.get_remote_keys().

    Useful to see changes in the keys within a bucket after some operation.
    Ex:
        before = list(bucket.get_remote_keys())
        do_something()
        diff = key_diff(before, bucket.get_remote_keys())

    Returned dict has fields 'new', 'removed', 'modified', and 'unmodified',
    and each field contains a list of str key names. Size and md5 are used to
    check for modification.

    Note: Be sure that 'before' is a list and *not* the generator returned
    by Bucket.get_remote_keys; because it's a generator, it won't actually
    fetch them from the server until they are iterated through.

    """
    before_set = {k['name'] for k in before}
    after_set = {k['name'] for k in after}

    before_keys = {k['name']: k for k in before if k['name'] in after_set}
    after_keys = {k['name']: k for k in after if k['name'] in before_set}

    removed = list(before_set-after_set)
    new = list(after_set-before_set)

    modified = []
    unmodified = []
    for k in (before_set & after_set):
        if before_keys[k]['md5'] == after_keys[k]['md5'] and \
           before_keys[k]['size'] == after_keys[k]['size']:
            unmodified.append(k)
        else:
            modified.append(k)
    return {'new': new, 'removed': removed, 'modified': modified,
            'unmodified': unmodified}


import logging
import os

import s3tup.utils as utils

log = logging.getLogger('s3tup.rsync')

class Rsync(object):
    def __init__(self, src=None, dest=None, delete=False, matcher=None):
        self.src = src
        self.dest = dest
        self.delete = delete
        self.matcher = matcher

    def run(self, bucket, protect=None):
        # Silence key logging to avoid redundant messages
        key_log = logging.getLogger('s3tup.key')
        key_log.setLevel(logging.WARNING)

        src = self.src or ''
        dest = self.dest or ''

        bucket_url = bucket.name+'.s3.amazonaws.com'
        log.info("rsyncing '{}'".format(os.path.join(bucket_url, dest)))
        log.info("    with '{}'...".format(os.path.abspath(src)))

        local_keys = self._get_local_keys()
        remote_keys = list(bucket.get_remote_keys(prefix=self.dest))

        # Is it smart to depend on key_diff for this?
        keys = utils.key_diff(remote_keys, local_keys)
        new_keys = keys['new']
        removed_keys = keys['removed']
        modified_keys = keys['modified']
        unmodified_keys = keys['unmodified']

        # Delete removed keys if delete is True
        if self.delete:
            bucket.delete_remote_keys(removed_keys)
        else:
            unmodified_keys.extend(removed_keys)
            removed_keys = []

        # Upload new keys
        for k in new_keys:
            key = bucket.make_key(k)
            log.info("new: '{}', uploading now...".format(k))
            self.rsync_key(key)

        # Upload modified keys
        for k in modified_keys:
            key = bucket.make_key(k)
            log.info("modified: '{}', uploading now...".format(k))
            self.rsync_key(key)

        # Unsilence key logging
        key_log.setLevel(logging.DEBUG)

        log.info('rsync complete!')
        log.info('{} new, {} removed, {} modified, {} unmodified'.format(
                 len(new_keys), len(removed_keys), len(modified_keys),
                 len(unmodified_keys)))

        return {'new': new_keys, 
                'removed': removed_keys, 
                'modified': modified_keys, 
                'unmodified': unmodified_keys}

    def rsync_key(self, key):
        """Rsyncs s3tup.key.Key 'key' with its corresponding local file.

        Takes a s3tup.key.Key object and rsyncs it with its corresponding
        local file. The local file path is basically just the key name minus
        self.dest relative to self.src.

        """
        src = self.src or ''
        dest = self.dest or ''
        unrelative = os.path.relpath(key.name, dest)
        local_path = os.path.join(src, unrelative)
        key.rsync(open(local_path, 'rb'))

    def _get_local_keys(self):
        """Return bucket.get_remote_keys-like response from files in self.src

        Uses utils.os_walk_relative to get all filenames relative to
        self.src, and then returns a list of dicts similar to
        Bucket.get_remote_keys. This value is used as an input to key_diff.

        """
        src = self.src or '.'
        dest = self.dest or ''
        matcher = self.matcher or utils.Matcher()
        keys = []
        for relpath in utils.os_walk_relative(src):
            if not matcher.matches(relpath):
                continue

            key_name = os.path.join(dest, relpath)
            local_path = os.path.join(src, relpath)

            key = {}
            key['name'] = key_name
            key['md5'] = utils.file_md5(open(local_path, 'rb'))
            key['size'] = os.path.getsize(local_path)
            keys.append(key)
        return keys

    def _get_protect(self, protect):
        if self.protect is not None:
            if protect is not None:
                return protect + self.protect
            else:
                return self.protect
        elif protect is not None:
            return protect
        else:
            return utils.Matcher(ignore_patterns=['*',])